/*********************************************************************/
/* File: buffer.C                                   Date 4/7 -98     */
/* Based on popproxy version 1.3 by N.Oren                           */
/* Programmer: Jens Landgr                                          */
/*********************************************************************/

#include "buffer.h"
#include "popenrw.h"
#include "wgpopp.h"

extern int errno;

/******************************************************************/
/* Creates a buffer with given size.                              */
/* NB! May not be constructed directly, must be used trough one   */
/* of sockbufferClass or procbufferClass.                         */
/******************************************************************/
bufferClass::bufferClass(int size)
{
  _size = size;
  _buf = new char[_size];
  memset(_buf, 0, _size);
  _tail = _buf;
}

/******************************************************************/
/* Uses the socket sock as io-stream.                             */
/******************************************************************/
sockbufferClass::sockbufferClass(int sock, int size): bufferClass(size)
{
  _file = sock;
}

/******************************************************************/
/* Starts a new process by executing com and connects _file to    */
/* its stdin and stdout.                                          */
/******************************************************************/
procbufferClass::procbufferClass(const char *com, int size): 
  bufferClass(size)
{
  if ((_file = popenrw(com)) == -1)
    panic("Could not popen \"%s\"", com);
}

/******************************************************************/
/* Flushes the stream and deallocates memory.                     */
/******************************************************************/
sockbufferClass::~sockbufferClass()
{
  fsync(_file);
  shutdown(_file, 2);
  close(_file);
}

/******************************************************************/
/* Closes the process and deallocates memory.                     */ 
/******************************************************************/
procbufferClass::~procbufferClass()
{
  fsync(_file);
  pcloserw(_file);
}

/******************************************************************/
/* Destructor for bufferClass.                                    */
/******************************************************************/
bufferClass::~bufferClass()
{
  delete _buf;
}

/******************************************************************/
/* Gets a line from buffer, returning how many characters that    */
/*   was extracted (not the length of buf!)                       */
/* Pre: buf contains '\n'.                                        */
/******************************************************************/
int bufferClass::getline(char *buf)
{
  int i = 0;
  while (_buf[i] != '\n') {
    buf[i] = _buf[i];
    i++;
  }
  /* If the '\r' is missing it will work anyway. */
  if ((i > 0) && (_buf[i-1] == '\r'))
    buf[i-1] = '\0';
  else
    buf[i] = '\0';

  return i+1;
}

/******************************************************************/
/* Reads from the stream and puts it in the buffer, removing      */
/* trailing return and newline.                                   */
/* TODO: Should have other errormessages, ENOMEM and ESPIPE does  */
/* describe the nature of the error.                              */
/******************************************************************/
int bufferClass::readl(char *buf)
{
  int count;
  int tmplen;

  while(1) {
    //If not a full line of text have been read:
    if (strchr(_buf,'\n')==NULL) {
      if (_tail >= _buf+_size-1) { //Buffer full?
	errno = ENOMEM; //Not enough space, close to the truth. 
	return -1;
      }
      //Should I block or should I loop?
      count = read(_file, _tail, _buf+_size-_tail);
      if (count == -1) //Error while reading
	return -1;
      if (count == 0) { //End-of-file reached, should not happen!
	errno = ESPIPE; //What error should I use?
	return -1;
      }
      _tail += count;
      *_tail = 0x0; //Marks end of data.
    } else { //Else just return the next line in the _buf.
      tmplen = getline(buf);
      strcpy(_buf, _buf+tmplen);
      _tail -= tmplen;
      return strlen(buf);
    }
  }
}

/******************************************************************/
/* Writes a string to the stream, and adds return and newline.    */
/******************************************************************/
int bufferClass::writel(const char *buf)
{
  int stat;
  //Should I make _file blocking or should I 
  //loop until everything is written?
  //For now, just pray that everything works alright.
  stat = write(_file, buf, strlen(buf));
  if (stat != -1)
    stat = write(_file, "\r\n", 2);
  if (stat != -1)
    fsync(_file);
  if (stat == -1)
    return -1;
  return 1;
}
