/*********************************************************************/
/* File: clientserver.C                             Date 4/7 -98     */
/* Based on popproxy version 1.3 by N.Oren                           */
/* Programmer: Jens Landgr                                          */
/* History:                                                          */
/* 13/7 -98   Added _malfunction to avoid infinite loops, panic()    */
/*            sends error messages, so the Client or Server need     */
/*	      to know if they caused the error.                      */
/* 12/8 -98   Removed a bug in the getCode() functions. They didn't  */
/*            check if the code was longer than MAXCODELEN.          */
/* 15/8 -98   Changed structure radically to make the code less      */
/*            errorprone, boosted efficiency by adding checkCom and  */
/*            changing transfer to use it.                           */
/* 29/9 -98   Removed a serious bug in getCom() and optimized the    */
/*            code in getCom() and checkCom() by removing strncpy()  */
/*            operation.					     */
/*********************************************************************/

#include "clientserver.h"
#include "wgpopp.h"

static char *Commands[] = {"+OK", "-ERR", ".", "AUTH", "USER", "PASS", "TOP", 
			   "LIST", "RETR", "UIDL", "QUIT"};

void popstreamClass::init()
{
  _malfunction = 0;
}

popstreamClass::popstreamClass(bufferClass *sock)
{
  init();
  _sock = sock;
  strcpy(_id, "stream");
}

popstreamClass::popstreamClass(bufferClass *sock, char *id)
{
  init();
  _sock = sock;
  strncpy(_id, id, IDLEN);
}

popstreamClass::~popstreamClass()
{
}

/******************************************************************/
/* Returns the command in the buffer.                             */
/******************************************************************/
POP_Command popstreamClass::getCom() 
{
  int i;
  int comlen;

  if (_buf[MAXCOMLEN-1] == ' ')
    comlen = MAXCOMLEN-1;
  else
    comlen = MAXCOMLEN;

  i = 0;
  while (i < C_NONE) {
    if (strncasecmp(_buf, Commands[i], comlen) == 0)
      return (POP_Command)i;
    i++;
  }

  return C_NONE;
}

/******************************************************************/
/* Returns true if pc is the command in the buffer.               */
/* A bit faster than getCom, can be used in tight loops.          */
/******************************************************************/
boolean popstreamClass::checkCom(POP_Command pc) 
{
  int comlen;

  if (_buf[MAXCOMLEN-1] == ' ')
    comlen = MAXCOMLEN-1;
  else
    comlen = MAXCOMLEN;

  if (strncasecmp(_buf, Commands[pc], comlen) == 0)
      return TRUE;

  return FALSE;
}

/******************************************************************/
/* Returns the data in the package, NULL if there is no data.     */
/* NB! The returned pointer points into the internal              */
/* representation. Dont deallocate!                              */
/******************************************************************/
char *popstreamClass::getData()
{
  char *data = _buf;
  while ((*data != ' ') && (*data != '\0'))
    data++;
  if (*data == ' ') data++; //Skip the blank.
  return (*data == '\0') ? (char*)NULL : data;
}

/******************************************************************/
/* Precondition: The current buffer contains the USER-signal.     */
/* Returns username as a copy, deallocate when done!              */
/******************************************************************/
char *popstreamClass::extractUname()
{
  char *uname = new char[MAXUNAME+1];
  char *data = getData();
  int i = 0;

  if (data == NULL)
    return NULL; //Probably not used correctly

  while ((data[i] != SEP_SYMBOL) && (data[i] != '\0')) {
    uname[i] = data[i];
    i++;
  }
  uname[i] = '\0';
  return uname;
}

/******************************************************************/
/* Precondition: The current buffer contains the USER-signal.     */
/* Postcondition: If a servername is found it is returned else    */
/*                NULL.                                           */
/* Returns servername as a copy.                                  */
/******************************************************************/
char *popstreamClass::extractServname()
{
  char *servname = new char[MAXSERVNAME+1];
  int i = 0;

  //Find the separator.
  while ((_buf[i] != SEP_SYMBOL) && (_buf[i] != '\0'))
    i++;
  if (_buf[i] == '\0')
    return (char*)NULL;
  i++;

  //Copy the servername
  strncpy(servname, &_buf[i], MAXSERVNAME);

  return servname;  
}

/******************************************************************/
/* Reads from stream into the internal buffer.                    */
/* NB! The returned pointer points at the internal buffer.        */
/******************************************************************/
char *popstreamClass::read()
{
  set_alarm(TIMEOUT);
  if (_sock->readl(_buf) == -1)
    panic("Error reading %s:%s", _id, strerror(errno));
  set_alarm(0);
  //This debug message is nice as long as no big data transfers occur.
  //But its only used during debugging anyway.
  DEBUG_MESS((LOG_INFO, "From %s: %s\n", _id, _buf));
  return _buf;
}

/******************************************************************/
/* Writes to stream                                               */
/******************************************************************/
char *popstreamClass::write(char *buf)
{
  if (_malfunction)
    return NULL;
  if (_sock->writel(buf) == -1) {
    _malfunction = 1;
    panic("Could not write to %s:%s", _id, strerror(errno));
  }
  DEBUG_MESS((LOG_INFO, "To %s: %s\n", _id, buf));
  return buf;
}

/******************************************************************/
/* Precondition: A multirow responce from a mailserver            */
/*               is expected.                                     */
/* Transfers a multirow responce from a mailserver.               */
/* If used when only a onerow responce is expected will get       */
/* a timeout.                                                     */
/******************************************************************/
void popstreamClass::transfer(popstreamClass &ps)
{
  write(ps.read());

  if (!ps.checkCom(C_ERR)) //Could the request be served?
    do {
      write(ps.read());
    } while (!ps.checkCom(C_END));
}

/******************************************************************/
/* Precondition: data is non-NULL.                                */
/* Sends a command to stream.                                     */
/******************************************************************/
void popstreamClass::sendCom(POP_Command com, char *data)
{
  char buf[BUFSIZE];
  snprintf(buf, BUFSIZE, "%s %s", Commands[com], data);
  write(buf);
}

/******************************************************************/
/* Sends a command to stream, without data.                       */
/******************************************************************/
void popstreamClass::sendCom(POP_Command com)
{
  write(Commands[com]);
}

//End of clientserver.C




