/*********************************************************************/
/* File: wgpopp.C                                   Date 4/7 -98     */
/* Based on popproxy version 1.3 by N.Oren                           */
/* Programmer: Jens Landgr                                          */
/* History:                                                          */
/* 13/7 -98   Had to add the _malfunction variable in Client/Server  */
/*            and a error signal in panic to avoid problems for      */
/*            client.                                                */
/*            Netscape has a bad habit of freezing while popping     */
/*            mail (doesn't netscape use timeout?)                   */
/* 11/8 -98   Added local popping trough popper and popenrw().       */
/*            Unfortunately I had to change popper to make this      */
/*            work, later I have to do something about this.         */
/* 4/9  -98   Added connectToPopper() to be able to connect to       */
/*            popper without changing popper. Instead users are      */
/*            required to add popper as a pop2 service.              */
/*********************************************************************/

#include "wgpopp.h"
#include "clientserver.h"
#include "popenrw.h"
#include <stdarg.h>

/***************************************************************/
/* Global variables.                                           */
/* It's pretty hard to avoid these.                            */
/***************************************************************/
popstreamClass *client = NULL;
popstreamClass *server = NULL;

#ifdef DEBUG
int debug = 0;
#endif

/***************************************************************/
/* Helpfunction.                                               */
/***************************************************************/
void shut_down(POP_Command com, char *mess)
{
  if (client != NULL) { //Must send an error to client if it didn't cause it.
    client->sendCom(com, mess);
    delete(client);
  }
  if (server != NULL) {//Shut down server nicely, if it didn't cause the error.
    server->sendCom(C_QUIT);
    delete(server);
  }
}

/***************************************************************/
/* panic()                                                     */
/* Works as printf but to the syslog, and shuts down.          */
/***************************************************************/
void panic(const char* format, ...)
{
  char errmess[100];
  va_list ap;

  va_start(ap, format);

  if (format != NULL) {
    vsnprintf(errmess, 100, format, ap); //Not all compilers support this!
    syslog(LOG_ERR, errmess);
  }
  va_end(ap);

  shut_down(C_ERR, "Abnormal exit");

  exit(ABNORMAL_EXIT);
}

/***************************************************************/
/* quit()                                                      */
/* Quits by sending signals in both directions.                */
/***************************************************************/
void quit()
{
  shut_down(C_OK, "bye ...");

  exit(NORMAL_EXIT);
}

/***************************************************************/
/* timeout()                                                   */
/* what to do if the alarm goes off                            */
/***************************************************************/
void timeout(int signal)
{
  syslog(LOG_ERR, "Timeout while reading client or server.\n");
  panic(NULL);
}

/***********************************************************/
/* SIGTERM and SIGKILL are handled here as follows:        */
/***********************************************************/
void die(int signal)
{
  panic(NULL);
}

/***************************************************************/
/* connectToMailserver()                                       */
/* Connects to the mailserver and returns a socket.            */
/***************************************************************/
int connectToMailserver(char *servname)
{
  hostent *hp;
  sockaddr_in sa;
  int hubsock;

  DEBUG_MESS((LOG_INFO, "Tries to connect to \"%s\"\n", servname));
  hp = gethostbyname(servname);
  if (hp == NULL)
    panic("Could not lookup pop host!\n");

  sa.sin_family = PF_INET;
  sa.sin_port = htons(HUBPORT);
  sa.sin_addr = *(in_addr *)hp->h_addr;

  hubsock = socket(AF_INET,SOCK_STREAM,0);
  if (connect(hubsock,(sockaddr *)&sa,sizeof(sa))==-1)
    panic("Unable to connect to pop host!!\n");

  return hubsock;
}

/***************************************************************/
/* connectToPopper()                                           */
/* Connects to the local popper and returns a socket.          */
/***************************************************************/
int connectToPopper()
{
  struct servent *sp;
  hostent *hp;
  sockaddr_in sa;
  int hubsock;

  sp = getservbyname(POPPERSERVICE, POPPERPROT);
  if (sp == NULL)
    panic("Local popping not enabled at %s/%s.", POPPERPROT, 
	  POPPERSERVICE);

  DEBUG_MESS((LOG_INFO, "Tries to connect to %s:%d", LOCALHOST,
	      sp->s_port));

  hp = gethostbyname(LOCALHOST);
  if (hp == NULL)
    panic("Could not lookup localhost!\n");

  sa.sin_family = PF_INET;
  sa.sin_port = sp->s_port;
  sa.sin_addr = *(in_addr *)hp->h_addr;

  hubsock = socket(AF_INET,SOCK_STREAM,0);
  if (connect(hubsock,(sockaddr *)&sa,sizeof(sa))==-1)
    panic("Unable to connect to %s:%d\n", LOCALHOST, sp->s_port);

  return hubsock;
}

/***************************************************************/
/* handleUSERsig()                                             */
/* Precondition: The USER-signal has been read from client.    */
/* Connects and sends the USER signal and handles the responce.*/
/***************************************************************/
void handleUSERsig()
{
  char *username = client->extractUname();
  char *servname = client->extractServname();
  bufferClass *bufcl;

  DEBUG_MESS((LOG_INFO, "User \"%s\" is connecting.", username));

  if (servname != NULL)
    bufcl =  new sockbufferClass(connectToMailserver(servname), BUFSIZE);
  else
    bufcl =  new sockbufferClass(connectToPopper(), BUFSIZE);
    //bufcl =  new procbufferClass(POPPER_COMLINE, BUFSIZE);

  server = new popstreamClass(bufcl, "server");

  server->read(); //Must get the startup-message.

  server->sendCom(C_USER, username);

  client->write(server->read()); //Gets the responce on USER-sig. 

  /* Username and servname are copies, need explicit deallocation. */
  delete(username);
  delete(servname);
}

/***************************************************************/
/* connectFase()                                             */
/* Precondition: The client is connected to.                   */
/* Postcondition: The server is connected to.                  */
/* Waits for USER-signal and connects to the mail-server.      */
/***************************************************************/
void connectFase()
{
  client->sendCom(C_OK, FIRSTSIGNAL);

  while (1) {
    client->read();  //Reads the next signal from client.
    switch (client->getCom()) {
    case C_AUTH:
      client->sendCom(AUTH_SIG, AUTH_RESPONCE); // Should I send this?
      break;
    case C_QUIT:
      quit();           //This exits the program.
      break;
    case C_USER:
      handleUSERsig();  //Takes care of connecting to server (if possible).
      return;
      break;
    default:
      DEBUG_MESS((LOG_ERR, "Got an unexpected command while connecting.\n"));
      client->sendCom(C_ERR, "huh???");
      break;
    } /* switch () */
  } /* while(1) */
}

/***************************************************************/
/* dataFase()                                                  */
/* Precondition: The server and client is connected.           */
/* Transfers data between client and server.                   */
/***************************************************************/
void dataFase()
{
  //Note! buffer points directly into the internal data of client.
  char *buffer; 
  int multi;

  while(1) {
    multi = 0;
    buffer = client->read();
    switch (client->getCom()) {
    case C_QUIT: 
      quit();
      break;
    case C_RETR:
    case C_TOP: 
      multi = 1;
      break;
    case C_LIST: //These two have different responce depending on parameters. 
    case C_UIDL: 
      if (client->getData() == NULL) //No arguments = multirow responce.
	multi = 1; 
      break;
    } /* switch () */
    
    server->write(buffer);
      
    if (multi == 0) // is it a multirow signal?
      client->write(server->read()); // Transfers one row.
    else
      client->transfer(*server); // Transfers multirows.
  } /* while (1) */
}

/***************************************************************/
/* handleArgs()                                                */
/* Just checking the debug option for the moment.              */
/***************************************************************/
void handleArgs(int argc, char **argv)
{
#ifdef DEBUG
#define NO_OF_ARGS 3
  struct { char *arg; 
    int argtype; } Args[] = {{"debug", BASIC_DEBUG}, 
			   {"dbgio", IO_DEBUG},
			   {"dbtime", TM_DEBUG}};
  int i = 0;
  int status = 1;

  if (argc == 2) {
    while ((i < NO_OF_ARGS) && (strcmp(argv[1], Args[i].arg) != 0)) 
      i++;
    if (i < NO_OF_ARGS)
      debug = Args[i].argtype;
    else 
      status = 0;
  } else
    if (argc > 2)
      status = 0;

  if (!status) {
    printf("Usage: %s [debug|dbgio|dbtime]\n", PROGRAM_NAME);
    exit(ABNORMAL_EXIT);
  }
#else
  if (argc > 1) {
    printf("%s does not take any arguments.\n", PROGRAM_NAME);
    exit(ABNORMAL_EXIT);
  }
#endif
}

/***************************************************************/
/* main()                                                      */
/* Sets some signals, opens log and starts the comunication.   */
/***************************************************************/
void main(int argc, char **argv)
{
  handleArgs(argc, argv);

  signal(SIGALRM,timeout);
  signal(SIGHUP,SIG_IGN);
  signal(SIGTERM,die);
  signal(SIGKILL,die);
  signal(SIGINT,SIG_IGN);
  signal(SIGPIPE,SIG_IGN);

  /* Open logging */
  openlog(PROGRAM_NAME,LOG_PID,LOG_DAEMON);

  DEBUG_MESS((LOG_INFO, "Starting up..."));

  //The socket pointer to the client is 0 (inetd connects it there).
  client = new popstreamClass(new sockbufferClass(0, BUFSIZE), "client");

  connectFase();
  dataFase();

  closelog();
}

